﻿using System;
using System.Collections.Generic;
using System.Text;

namespace opentkprojekt.Interpolation
{
    class CatmullRomInterpolator : Interpolator
    {
        public KeyValuePair<KeyFraming.ObjectKeyFrame, KeyFraming.CameraKeyFrame> Interpolate(int time, ref KeyFraming.KeyFrameArray kfa)
        {
            if (kfa.IsKeyframeInTime(time))
            {
                return kfa.GetKeyframeByTime(time);
            }

            int beforeFirst = 0;
            int first = 0;
            int second = 0;
            int afterSecond = 0;
            kfa.FindKeyframesIndexes(time, out first, out second);
            if (first > 0) { beforeFirst = first - 1; } else { beforeFirst = first; }
            if (second < kfa.GetKeyframeIndexesCound() - 1) { afterSecond = second + 1; } else { afterSecond = second; }
            int timeFirst = kfa.GetKeyframeTimeByIndex(first);
            int timeSecond = kfa.GetKeyframeTimeByIndex(second);

            double dtime = (double)(time - timeFirst) / (double)(timeSecond - timeFirst);
            if (Double.IsNaN(dtime)) { dtime = 0; }

            KeyFraming.CameraKeyFrame ckf = new KeyFraming.CameraKeyFrame();
            KeyFraming.ObjectKeyFrame okf = new KeyFraming.ObjectKeyFrame();

            var kf0 = kfa.GetKeyframeByIndex(beforeFirst);
            var kf1 = kfa.GetKeyframeByIndex(first);
            var kf2 = kfa.GetKeyframeByIndex(second);
            var kf3 = kfa.GetKeyframeByIndex(afterSecond);

            ckf.SetFollowTarget(kf2.Value.GetFollowTarget());
            ckf.SetPosition(InterpolateIt(dtime, kf0.Value.GetPosition(), kf1.Value.GetPosition(), kf2.Value.GetPosition(), kf3.Value.GetPosition()));
            var ct0 = kf0.Value.GetFollowTarget() ? kf0.Key.GetTranslation() : kf0.Value.GetTarget();
            var ct1 = kf1.Value.GetFollowTarget() ? kf1.Key.GetTranslation() : kf1.Value.GetTarget();
            var ct2 = kf2.Value.GetFollowTarget() ? kf2.Key.GetTranslation() : kf2.Value.GetTarget();
            var ct3 = kf3.Value.GetFollowTarget() ? kf3.Key.GetTranslation() : kf3.Value.GetTarget();
            ckf.SetTarget(InterpolateIt(dtime, ct0, ct1, ct2, ct3));

            okf.SetTranslation(InterpolateIt(dtime, kf0.Key.GetTranslation(), kf1.Key.GetTranslation(), kf2.Key.GetTranslation(), kf3.Key.GetTranslation()));
            okf.SetScale(InterpolateIt(dtime, kf0.Key.GetScale(), kf1.Key.GetScale(), kf2.Key.GetScale(), kf3.Key.GetScale()));
            okf.SetRotation(InterpolateIt(dtime, kf0.Key.GetRotationQuaternion(), kf1.Key.GetRotationQuaternion(), kf2.Key.GetRotationQuaternion(), kf3.Key.GetRotationQuaternion()));

            return new KeyValuePair<KeyFraming.ObjectKeyFrame, KeyFraming.CameraKeyFrame>(okf, ckf);
        }

        private OpenTK.Vector3d InterpolateIt(double time, OpenTK.Vector3d v0, OpenTK.Vector3d v1, OpenTK.Vector3d v2, OpenTK.Vector3d v3)
        {
            double time10 = time + 1.0;
            double time12 = time - 1.0;
            var v10 = (1 - time10) * v0 + time10 * v1;
            var v11 = (1 - time) * v1 + time * v2;
            var v12 = (1 - time12) * v2 + time12 * v3;
            double time20 = (time + 1.0) / 2.0;
            double time21 = time / 2.0;
            var v20 = (1 - time20) * v10 + time20 * v11;
            var v21 = (1 - time21) * v11 + time21 * v12;
            return (1 - time) * v20 + time * v21;
        }

        private OpenTK.Quaterniond InterpolateIt(double time, OpenTK.Quaterniond q0, OpenTK.Quaterniond q1, OpenTK.Quaterniond q2, OpenTK.Quaterniond q3)
        {
            double time10 = time + 1.0;
            double time12 = time - 1.0;
            var q10 = SphericalLinearInterpolation.Interpolate(q0, q1, time10);
            var q11 = SphericalLinearInterpolation.Interpolate(q1, q2, time);
            var q12 = SphericalLinearInterpolation.Interpolate(q2, q3, time12);
            double time20 = (time + 1.0) / 2.0;
            double time21 = time / 2.0;
            var q20 = SphericalLinearInterpolation.Interpolate(q10, q11, time20);
            var q21 = SphericalLinearInterpolation.Interpolate(q11, q12, time21);
            return SphericalLinearInterpolation.Interpolate(q20, q21, time);
        }


        public string InterpolatorName()
        {
            return "Catmull-Rom";
        }
    }
}
